/*
 * An implementation of integer factorization algorithms
 * See http://www.giaithuatlaptrinh.com/?p=34 for details
 *
 *  Created on: Aug 5, 2015
 *      Author: hunglv
 */

#include<time.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#define PRIME 1
#define COMPOSITE 0
#define MAXNUMFACT 20 // the maximum number of factors

typedef unsigned long long int LLU; // using unsigned long long type to avoid overflow

LLU Fa[MAXNUMFACT];// the array storing factors
int nfact = -1// the number of factors

LLU gcd(LLU a, LLU b);
int witness(LLU a,  LLU N);
void decompose(LLU p, int *k, LLU *m);
LLU random64();
LLU mod_mul(LLU a, LLU b, LLU p);
LLU mod_power(LLU a, LLU b, LLU p);
int miller_rabin_testing(int acc, LLU N);
void fast_native_fact(LLU N);
void print_factors(void);
LLU llu_fsqrt(LLU N);

int main (void){
	fast_native_fact((LLU)9383); // 11*853
	print_factors();
	return 0;
}

/*
 * Q[0,1,...] = {2,3,5,7,11,13,17,19,23,25,29,31,35,....}
 * Except for 2 and 3, the difference between two consecutive elements of Q is 2 or 4 alternatively.
 */

void fast_native_fact(LLU N){
	if(N == 1){
		return;
	}
	while (N % 2 == 0){
		Fa[++nfact] = (LLU)2;
		N = N/2;
	}
	while (N % 3 == 0){
		Fa[++nfact] = (LLU)3;
		N = N/3;
	}
	if( miller_rabin_testing(10, N) == PRIME ){
			Fa[++nfact] = N;
			return;
	}
	LLU Q = 5;
	int diff = 2;
	int sqrtN = llu_fsqrt(N);
	while(Q <= sqrtN){
		if(N %Q == 0){
			Fa[++nfact] = Q;
			N = N/Q;
			sqrtN = llu_fsqrt(N);
			if(miller_rabin_testing(10, N) == PRIME ){
				Fa[++nfact] = N;
				return;
			}
		}else {
			Q = Q + (LLU)diff;
			diff = (diff&2) + 2;// diff is alternating between 2 and 4
		}
	}
}

void print_factors(){
	int i = 0;
	for(i = 0; i < nfact-1; i++){
		printf("%llu*", Fa[i]);
	}
	printf("%llu\n", Fa[nfact-1]);
}

/*
 * From here to the rest is an implementation of Miller-Rabin primality testing algorithm
 * See http://www.giaithuatlaptrinh.com/?p=278 for details
 */

int miller_rabin_testing(int acc, LLU N){
	LLU a = 0;
	int i = 0;
	for(i = 0; i <= acc; i++){
		a = random64()%(N-2)+ 2;
		if(gcd(a,N) != 1){
			return COMPOSITE;
		} else if (mod_power(a,N-1,N) != 1){
			return COMPOSITE;
		}else {
			if (witness(a,N) == COMPOSITE){
				return COMPOSITE;
			}
		}
	}
	return PRIME;
}

LLU gcd(LLU a, LLU b){
	LLU tmp;
	while  (b != 0){
		tmp  = b;
		b = a % b;
		a = tmp;
	}
	return a;
}

LLU mod_power(LLU a, LLU b, LLU p){
	if (b == 1){
		return a %p;
	} else {
		LLU x = mod_power(a, b/2, p);
		if((b& 1) == 0){ // b is even
			return mod_mul(x,x,p); // return  (x*x)%p;
		} else {
			return (mod_mul(mod_mul(x,x,p), a, p));// return (x*x*a)%p;
		}
	}
}
/*
 * this function compute (a*b) \mod p using (\log_2a + 1) space
 */

LLU mod_mul(LLU a, LLU b, LLU p){
	if (b == 1){
		return a %p;
	} else {
		LLU x = mod_mul(a, b/2, p);
		if((b& 1) == 0){ // b is even
			return (x+x)%p;
		} else {
			return (((x+x)%p)+a)%p;
		}
	}
}


int witness(LLU a,  LLU N){
	LLU m;
	int k;
	decompose(N-1,&k,&m);
	LLU B[k+1];
	B[0] = mod_power(a,m,N);
	if(B[0] == 1){
		return PRIME;
	} else {
		int i = 1;
		while(i <= k){
			B[i] =  mod_mul(B[i-1],B[i-1],N);// (B[i-1]*B[i-1])%N;
			if(B[i] == 1){
				if(B[i-1] == N-1){
					return PRIME;
				} else {
					return COMPOSITE;
				}
			}
			i++;
		}
	}
	return COMPOSITE;
}

void decompose(LLU p, int *k, LLU *m){
	int i = 0;
	while ((p& 1) == 0){ // p is even
		i++;
		p = p >> 1; // p = p /2
	}
	*k = i;
	*m = p;
}
/*
 * An implementation of the function that returns the floor of the square root of an integer
 * See http://www.giaithuatlaptrinh.com/?p=402 for details
 */

LLU llu_fsqrt(LLU  N){
	LLU X, A2, R;
	X = N;
	A2 = ((LLU)1) << 62; 
	R = 0;
	while(N < A2){
		A2 >>= 2;
	}

	while(A2 > 0){
		if (X >= R + A2) {
			X -= R + A2;
			R = (R >> 1) + A2;
			A2 = A2>> 2;
		} else {
			A2 = A2 >> 2;
			R = R >> 1;
		}
	}
	return R;
}

